from collections import namedtuple

Field = namedtuple('Field', ('name', 'data'))
Downloads = Field('download_count', 'COUNT(*)')
Date = Field('download_date', 'FORMAT_TIMESTAMP("%Y-%m-%d", timestamp)')
Month = Field('download_month', 'FORMAT_TIMESTAMP("%Y-%m", timestamp)')
Year = Field('download_year', 'FORMAT_TIMESTAMP("%Y", timestamp)')
Country = Field('country', 'country_code')
Project = Field('project', 'file.project')
Version = Field('version', 'file.version')
File = Field('file', 'file.filename')
PythonVersion = Field('python_version', r'REGEXP_EXTRACT(details.python, r"^([^\.]+\.[^\.]+)")')
Percent3 = Field(
    'percent_3',
    r'ROUND(100 * SUM(CASE WHEN REGEXP_EXTRACT(details.python, r"^([^\.]+)") = "3" THEN 1 ELSE 0 END) / COUNT(*), 1)',
)
Percent2 = Field(
    'percent_2',
    r'ROUND(100 * SUM(CASE WHEN REGEXP_EXTRACT(details.python, r"^([^\.]+)") = "2" THEN 1 ELSE 0 END) / COUNT(*), 1)',
)
Implementation = Field('implementation', 'details.implementation.name')
ImplementationVersion = Field('impl_version', r'REGEXP_EXTRACT(details.implementation.version, r"^([^\.]+\.[^\.]+)")')
OpenSSLVersion = Field('openssl_version', 'REGEXP_EXTRACT(details.openssl_version, r"^OpenSSL ([^ ]+) ")')
Installer = Field('installer_name', 'details.installer.name')
InstallerVersion = Field('installer_version', 'details.installer.version')
SetuptoolsVersion = Field('setuptools_version', 'details.setuptools_version')
System = Field('system_name', 'details.system.name')
SystemRelease = Field('system_release', 'details.system.release')
Distro = Field('distro_name', 'details.distro.name')
DistroVersion = Field('distro_version', 'details.distro.version')
CI = Field('ci', 'details.ci')
CPU = Field('cpu', 'details.cpu')
Libc = Field('libc_name', 'details.distro.libc.lib')
LibcVersion = Field('libc_version', 'details.distro.libc.version')

AGGREGATES = {Downloads, Percent3, Percent2}



from freezegun import freeze_time
from typing import Any, Iterator, List, Tuple
import copy
import pytest
import re

from packaging.specifiers import Specifier
from packaging.version import Version

from google.cloud.bigquery.schema import SchemaField
from google.cloud.bigquery.table import RowIterator

from pypinfo import core
from pypinfo.fields import Field, File, Percent3, PythonVersion

ROWS = [
    ['python_version', 'percent', 'download_count'],
    ['2.7', '51.7%', '342250'],
    ['3.6', '21.1%', '139745'],
    ['3.5', '17.2%', '114254'],
    ['3.4', '7.6%', '50584'],
    ['3.3', '1.0%', '6666'],
    ['3.7', '0.7%', '4516'],
    ['2.6', '0.7%', '4451'],
    ['3.2', '0.0%', '138'],
    ['None', '0.0%', '13'],
]


def test_create_config() -> None:
    # Act
    config = core.create_config()

    # Assert
    assert not config.use_legacy_sql


def test_normalize_dates_yyy_mm() -> None:
    # Arrange
    start_date = "2019-03"
    end_date = "2019-03"

    # Act
    start_date, end_date = core.normalize_dates(start_date, end_date)

    # Assert
    assert start_date == "2019-03-01"
    assert end_date == "2019-03-31"


def test_normalize_dates_yyy_mm_dd_and_negative_integer() -> None:
    # Arrange
    start_date = "2019-03-18"
    end_date = "-1"

    # Act
    start_date, end_date = core.normalize_dates(start_date, end_date)

    # Assert
    assert start_date == "2019-03-18"
    assert end_date == "-1"


def test_create_client_file_is_none() -> None:
    # Act / Assert
    with pytest.raises(SystemError):
        core.create_client(None)


def test_create_client_with_filename() -> None:
    # Arrange
    filename = "tests/data/sample-credentials.json"

    # Act
    output = core.create_client(filename)

    # Assert
    assert output.project == "pypinfo-test"


@pytest.mark.parametrize("test_input", ["-1", "2018-05-15"])
def test_validate_date_valid(test_input: str) -> None:
    # Act
    valid = core.validate_date(test_input)

    # Assert
    assert valid


@pytest.mark.parametrize("test_input", ["1", "2018-19-39", "something invalid"])
def test_validate_date_invalid(test_input: str) -> None:
    # Act / Assert
    with pytest.raises(ValueError):
        core.validate_date(test_input)


def test_format_date_negative_number() -> None:
    # Arrange
    dummy_format = "dummy format {}"

    # Act
    date = core.format_date("-1", dummy_format)

    # Assert
    assert date == 'TIMESTAMP_ADD(CURRENT_TIMESTAMP(), INTERVAL -1 DAY)'


def test_format_date_yyy_mm_dd() -> None:
    # Act
    date = core.format_date("2018-05-15", core.START_TIMESTAMP)

    # Assert
    assert date == 'TIMESTAMP("2018-05-15 00:00:00")'


def test_month_yyyy_mm() -> None:
    # Act
    first, last = core.month_ends("2019-03")

    # Assert
    assert first == "2019-03-01"
    assert last == "2019-03-31"


def test_month_yyyy_mm_dd() -> None:
    # Act / Assert
    with pytest.raises(ValueError):
        core.month_ends("2019-03-18")


def test_month_negative_integer() -> None:
    # Act / Assert
    with pytest.raises(ValueError):
        core.month_ends("-30")


def test_build_query() -> None:
    # Arrange
    # Data from pycodestyle in 2017-10
    # pypinfo -sd 2017-10-01 -ed 2017-10-31 -pc -l 100 --json pycodestyle pyversion
    project = "pycodestyle"
    all_fields = [PythonVersion]
    start_date = "2017-10-01"
    end_date = "2017-10-31"
    days = None
    limit = 100
    where = None
    order = None
    pip = True
    expected = r"""
SELECT
  REGEXP_EXTRACT(details.python, r"^([^\.]+\.[^\.]+)") as python_version,
  COUNT(*) as download_count,
FROM `bigquery-public-data.pypi.file_downloads`
WHERE timestamp BETWEEN TIMESTAMP("2017-10-01 00:00:00") AND TIMESTAMP("2017-10-31 23:59:59")
  AND file.project = "pycodestyle"
  AND details.installer.name = "pip"
GROUP BY
  python_version
ORDER BY
  download_count DESC
LIMIT 100
    """.strip()

    # Act
    output = core.build_query(project, all_fields, start_date, end_date, days, limit, where, order, pip)

    # Assert
    assert output == expected


def test_build_query_specifier() -> None:
    # pypinfo -sd -2 -ed -1 -l 20  --test 'foo==1'
    project = "foo==1"
    all_fields: List[Field] = []
    start_date = "-2"
    end_date = "-1"
    days = None
    limit = 20
    where = None
    order = None
    pip = True
    expected = r"""
SELECT
  COUNT(*) as download_count,
FROM `bigquery-public-data.pypi.file_downloads`
WHERE timestamp BETWEEN TIMESTAMP_ADD(CURRENT_TIMESTAMP(), INTERVAL -2 DAY) AND TIMESTAMP_ADD(CURRENT_TIMESTAMP(), INTERVAL -1 DAY)
  AND file.project = "foo"
  AND REGEXP_CONTAINS(file.version, r"(?i)^(0+!)?0*1(\.0+)*$")
  AND details.installer.name = "pip"
ORDER BY
  download_count DESC
LIMIT 20
        """.strip()  # noqa: E501

    # Act
    output = core.build_query(project, all_fields, start_date, end_date, days, limit, where, order, pip)

    # Assert
    assert output == expected


def test_build_query_days() -> None:
    # Arrange
    # Data from pycodestyle in 2017-10
    # pypinfo -sd 2017-10-01 -ed 2017-10-31 -pc -l 100 --json pycodestyle pyversion
    project = "pycodestyle"
    all_fields = [PythonVersion]
    start_date = None
    end_date = None
    days = 10
    limit = 100
    where = None
    order = None
    pip = True
    expected = r"""
SELECT
  REGEXP_EXTRACT(details.python, r"^([^\.]+\.[^\.]+)") as python_version,
  COUNT(*) as download_count,
FROM `bigquery-public-data.pypi.file_downloads`
WHERE timestamp BETWEEN TIMESTAMP_ADD(CURRENT_TIMESTAMP(), INTERVAL -11 DAY) AND TIMESTAMP_ADD(CURRENT_TIMESTAMP(), INTERVAL -1 DAY)
  AND file.project = "pycodestyle"
  AND details.installer.name = "pip"
GROUP BY
  python_version
ORDER BY
  download_count DESC
LIMIT 100
    """.strip()  # noqa: E501

    # Act
    output = core.build_query(project, all_fields, start_date, end_date, days, limit, where, order, pip)

    # Assert
    assert output == expected


def test_build_query_where() -> None:
    # Arrange
    # pypinfo -sd -2 -ed -1 --test --where 'file.filename LIKE "%manylinux%"' numpy file
    project = "numpy"
    all_fields = [File]
    start_date = "-2"
    end_date = "-1"
    days = None
    limit = 10
    where = 'file.filename LIKE "%manylinux%"'
    order = None
    pip = True
    expected = r"""
SELECT
  file.filename as file,
  COUNT(*) as download_count,
FROM `bigquery-public-data.pypi.file_downloads`
WHERE timestamp BETWEEN TIMESTAMP_ADD(CURRENT_TIMESTAMP(), INTERVAL -2 DAY) AND TIMESTAMP_ADD(CURRENT_TIMESTAMP(), INTERVAL -1 DAY)
  AND file.project = "numpy"
  AND details.installer.name = "pip"
  AND file.filename LIKE "%manylinux%"
GROUP BY
  file
ORDER BY
  download_count DESC
LIMIT 10
    """.strip()  # noqa: E501

    # Act
    output = core.build_query(project, all_fields, start_date, end_date, days, limit, where, order, pip)

    # Assert
    assert output == expected


def test_build_query_no_project() -> None:
    # pypinfo -sd -2 -ed -1 -l 20 --all --test ''
    project = ""
    all_fields: List[Field] = []
    start_date = "-2"
    end_date = "-1"
    days = None
    limit = 20
    where = None
    order = None
    pip = False
    expected = r"""
SELECT
  COUNT(*) as download_count,
FROM `bigquery-public-data.pypi.file_downloads`
WHERE timestamp BETWEEN TIMESTAMP_ADD(CURRENT_TIMESTAMP(), INTERVAL -2 DAY) AND TIMESTAMP_ADD(CURRENT_TIMESTAMP(), INTERVAL -1 DAY)
ORDER BY
  download_count DESC
LIMIT 20
        """.strip()  # noqa: E501

    # Act
    output = core.build_query(project, all_fields, start_date, end_date, days, limit, where, order, pip)

    # Assert
    assert output == expected


def test_build_query_only_aggregate() -> None:
    # pypinfo -sd -2 -ed -1 -l 20 numpy percent3
    project = "numpy"
    all_fields = [Percent3]
    start_date = "-2"
    end_date = "-1"
    days = None
    limit = 20
    where = None
    order = None
    pip = True
    expected = r"""
SELECT
  ROUND(100 * SUM(CASE WHEN REGEXP_EXTRACT(details.python, r"^([^\.]+)") = "3" THEN 1 ELSE 0 END) / COUNT(*), 1) as percent_3,
  COUNT(*) as download_count,
FROM `bigquery-public-data.pypi.file_downloads`
WHERE timestamp BETWEEN TIMESTAMP_ADD(CURRENT_TIMESTAMP(), INTERVAL -2 DAY) AND TIMESTAMP_ADD(CURRENT_TIMESTAMP(), INTERVAL -1 DAY)
  AND file.project = "numpy"
  AND details.installer.name = "pip"
ORDER BY
  download_count DESC
LIMIT 20
        """.strip()  # noqa: E501

    # Act
    output = core.build_query(project, all_fields, start_date, end_date, days, limit, where, order, pip)

    # Assert
    assert output == expected


def test_build_query_bad_end_date() -> None:
    # Arrange
    project = "pycodestyle"
    all_fields = [PythonVersion]
    # End date is before start date
    start_date = "-1"
    end_date = "-100"

    # Act / Assert
    with pytest.raises(ValueError):
        core.build_query(project, all_fields, start_date, end_date)


def test_build_query_bad_project_extras() -> None:
    with pytest.raises(ValueError, match=".*extras.*"):
        core.build_query('foo[bar]', [])


def test_build_query_bad_project_url() -> None:
    with pytest.raises(ValueError, match=".*url.*"):
        core.build_query('foo@https://foo.bar/', [])


def test_build_query_bad_project_marker() -> None:
    with pytest.raises(ValueError, match=".*marker.*"):
        core.build_query('foo ; sys_platform == "win32"', [])


def test_add_percentages() -> None:
    # Arrange
    rows = [
        ['python_version', 'download_count'],
        ['2.7', '480056'],
        ['3.6', '328008'],
        ['3.5', '149663'],
        ['3.4', '36837'],
        ['3.7', '1883'],
        ['2.6', '591'],
        ['3.3', '274'],
        ['3.2', '10'],
        ['None', '9'],
        ['3.8', '2'],
    ]

    expected = [
        ['python_version', 'percent', 'download_count'],
        ['2.7', '48.13%', '480056'],
        ['3.6', '32.89%', '328008'],
        ['3.5', '15.01%', '149663'],
        ['3.4', '3.69%', '36837'],
        ['3.7', '0.19%', '1883'],
        ['2.6', '0.06%', '591'],
        ['3.3', '0.03%', '274'],
        ['3.2', '0.00%', '10'],
        ['None', '0.00%', '9'],
        ['3.8', '0.00%', '2'],
    ]

    # Act
    with_percentages = core.add_percentages(rows)

    # Assert
    assert with_percentages == expected


def test_add_download_total() -> None:
    # Arrange
    rows = copy.deepcopy(ROWS)
    expected = copy.deepcopy(ROWS)
    expected.append(["Total", "", "662617"])

    # Act
    rows_with_total = core.add_download_total(rows)

    # Assert
    assert rows_with_total == expected


def test_tabulate_default() -> None:
    # Arrange
    rows = copy.deepcopy(ROWS)
    expected = """\
| python_version | percent | download_count |
| -------------- | ------- | -------------- |
| 2.7            |   51.7% |        342,250 |
| 3.6            |   21.1% |        139,745 |
| 3.5            |   17.2% |        114,254 |
| 3.4            |    7.6% |         50,584 |
| 3.3            |    1.0% |          6,666 |
| 3.7            |    0.7% |          4,516 |
| 2.6            |    0.7% |          4,451 |
| 3.2            |    0.0% |            138 |
| None           |    0.0% |             13 |
"""

    # Act
    tabulated = core.tabulate(rows)

    # Assert
    assert tabulated == expected


def test_tabulate_markdown() -> None:
    # Arrange
    rows = copy.deepcopy(ROWS)
    expected = """\
| python_version | percent | download_count |
| -------------- | ------: | -------------: |
| 2.7            |   51.7% |        342,250 |
| 3.6            |   21.1% |        139,745 |
| 3.5            |   17.2% |        114,254 |
| 3.4            |    7.6% |         50,584 |
| 3.3            |    1.0% |          6,666 |
| 3.7            |    0.7% |          4,516 |
| 2.6            |    0.7% |          4,451 |
| 3.2            |    0.0% |            138 |
| None           |    0.0% |             13 |
"""

    # Act
    tabulated = core.tabulate(rows, markdown=True)

    # Assert
    assert tabulated == expected


@freeze_time("2020-07-14 07:11:49")
def test_format_json() -> None:
    # Arrange
    # Data from pycodestyle in 2017-10
    # pypinfo -sd 2017-10-01 -ed 2017-10-31 -pc -l 100 --json pycodestyle pyversion
    rows = [
        ['python_version', 'percent', 'download_count'],
        ['2.7', '0.54', '587705'],
        ['3.6', '0.21', '227800'],
        ['3.5', '0.16', '169851'],
        ['3.4', '0.078', '84599'],
        ['3.3', '0.0091', '9953'],
        ['3.7', '0.0044', '4770'],
        ['2.6', '0.0041', '4476'],
        ['3.2', '0.0003', '326'],
        ['2.8', '3.7e-06', '4'],
        ['None', '2.8e-06', '3'],
    ]
    query_info = {
        'cached': False,
        'bytes_processed': 18087095002,
        'bytes_billed': 18087936000,
        'estimated_cost': '0.09',
    }
    indent = None
    expected = (
        '{"last_update":"2020-07-14 07:11:49",'
        '"query":'
        '{"bytes_billed":18087936000,'
        '"bytes_processed":18087095002,'
        '"cached":false,'
        '"estimated_cost":"0.09"},'
        '"rows":['
        '{"download_count":587705,"percent":"0.54","python_version":"2.7"},'
        '{"download_count":227800,"percent":"0.21","python_version":"3.6"},'
        '{"download_count":169851,"percent":"0.16","python_version":"3.5"},'
        '{"download_count":84599,"percent":"0.078","python_version":"3.4"},'
        '{"download_count":9953,"percent":"0.0091","python_version":"3.3"},'
        '{"download_count":4770,"percent":"0.0044","python_version":"3.7"},'
        '{"download_count":4476,"percent":"0.0041","python_version":"2.6"},'
        '{"download_count":326,"percent":"0.0003","python_version":"3.2"},'
        '{"download_count":4,"percent":"3.7e-06","python_version":"2.8"},'
        '{"download_count":3,"percent":"2.8e-06","python_version":"None"}]}'
    )

    # Act
    output = core.format_json(rows, query_info, indent)

    # Assert
    assert output == expected


def test_parse_query_result() -> None:
    data: List[Tuple[Any, ...]] = [
        ("name", "other"),
        ("name1", 1),
        ("name2", 2),
    ]
    expected = [[str(cell) for cell in row] for row in data]
    schema = (
        SchemaField(data[0][0], "STRING"),
        SchemaField(data[0][1], "INTEGER"),
    )

    class MockRowIterator(RowIterator):  # type: ignore[misc]
        def __init__(self) -> None:
            super().__init__(None, None, None, schema)

        def __iter__(self) -> Iterator[Tuple[Any, ...]]:
            return iter(data[1:])

    actual = core.parse_query_result(MockRowIterator())
    assert actual == expected


SPECIFIER_GOOD = {
    '==1!2.3.4': r'(?i)^0*1!0*2\.0*3\.0*4(\.0+)*$',
    '==1.0.0': r'(?i)^(0+!)?0*1(\.0+)*$',
    '==1-pre_post.dev': r'(?i)^(0+!)?0*1(\.0+)*[-_\.]?(c|rc|pre|preview)([-_\.]?0+)?(-0*0|[-_\.]?(post|rev|r)([-_\.]?0+)?)[-_\.]?dev([-_\.]?0+)?$',  # noqa: E501
    '==0a1-1.dev2': r'(?i)^(0+!)?0*0(\.0+)*[-_\.]?(a|alpha)[-_\.]?0*1(-0*1|[-_\.]?(post|rev|r)[-_\.]?0*1)[-_\.]?dev[-_\.]?0*2$',  # noqa: E501
    '==1.1.*': r'(?i)^(0+!)?(0*1\.0*1(\.0+)*|0*1\.0*1(\.[0-9]+)*)([-_\.]?(a|b|c|rc|alpha|beta|pre|preview)[-_\.]?[0-9]+?)?(-[0-9]+|([-_\.]?(post|rev|r)[-_\.]?[0-9]+?))?([-_\.]?dev[-_\.]?[0-9]+?)?$',  # noqa: E501
}


@pytest.mark.parametrize('specifier', SPECIFIER_GOOD.keys())
def test_specifier_good(specifier: str) -> None:
    actual = core.version_specifier_condition(Specifier(specifier)).strip()
    regex = SPECIFIER_GOOD[specifier]
    expected = f'REGEXP_CONTAINS(file.version, r"{regex}")'
    assert actual == expected


@pytest.mark.parametrize(
    'specifier, version',
    (
        ('==1!2.3.4', '1!2.3.4'),
        ('==1!2.3.4', '1!2.3.4.0'),
        ('==1!2.3.4', '01!02.03.04.00'),
        ('==1.0.0', '1.0.0'),
        ('==1.0.0', '1'),
        ('==1.0.0', '0!1.0'),
        ('==1.0.0', '00!01.00'),
        ('==1-pre_post.dev', '1-pre_post.dev'),
        ('==1-pre_post.dev', '1.0.0Rc0.poSt.0.dEv-0'),
        ('==1-pre_post.dev', '1.0.0_C00-Rev00_dEv.00'),
        ('==1-pre_post.dev', '1.0.0prevIew0-0-dEv00'),
        ('==1-pre_post.dev', '1.0.0prevIewR0dEv'),
        ('==1-pre_post.dev', '1.0.0rcRdEv'),
        ('==0a1-1.dev2', '0a1-1.dev2'),
        ('==0a1-1.dev2', '0.0.alpha.1.post.1.dev.2'),
        ('==0a1-1.dev2', '00!00.00.alpha.01.post.01.dev.02'),
        ('==1.1.*', '1.1'),
        ('==1.1.*', '1.1a1'),
        ('==1.1.*', '01.01.00.01.post01'),
        ('==1.1.*', '1.1.1.1.0.dev1'),
    ),
)
def test_specifier_regex_match(specifier: str, version: str) -> None:
    specifier_ = Specifier(specifier)
    version_ = Version(version)
    assert specifier_.contains(version_, prereleases=True)
    assert re.match(SPECIFIER_GOOD[specifier], str(version_))
    assert re.match(SPECIFIER_GOOD[specifier], version)


@pytest.mark.parametrize(
    'specifier, version',
    (
        ('==1!2.3.4', '1!2.3'),
        ('==1!2.3.4', '2.3.4'),
        ('==1!2.3.4', '1!2.3.4.post1'),
        ('==1!2.3.4', '1!2.3.4rc1'),
        ('==1!2.3.4', '1!2.3.4dev'),
        ('==1.0.0', '1.1'),
        ('==1.0.0', '1!1'),
        ('==1.0.0', '2'),
        ('==1.0.0', '1rc1'),
        ('==1-pre_post.dev', '1-pre1_post.dev'),
        ('==1-pre_post.dev', '1.0.0Rc0.poSt.1.dEv-0'),
        ('==1-pre_post.dev', '1.0.0_C00-Rev00_dEv.10'),
        ('==1-pre_post.dev', '1.0.0prevIew0-1-dEv00'),
        ('==1-pre_post.dev', '1.0.1prevIewR0dEv'),
        ('==1-pre_post.dev', '1!1.0.0rcRdEv'),
        ('==0a1-1.dev2', '0a1-1.dev3'),
        ('==0a1-1.dev2', '0.1.alpha.1.post.1.dev.2'),
        ('==0a1-1.dev2', '0!1.00.alpha.01.post.01.dev.02'),
        ('==1.1.*', '1.2'),
        ('==1.1.*', '1.0'),
        ('==1.1.*', '1!1.1'),
        ('==1.1.*', '2'),
    ),
)
def test_specifier_regex_no_match(specifier: str, version: str) -> None:
    specifier_ = Specifier(specifier)
    version_ = Version(version)
    assert not specifier_.contains(version_, prereleases=True)
    assert re.match(SPECIFIER_GOOD[specifier], str(Version(version))) is None
    assert re.match(SPECIFIER_GOOD[specifier], version) is None


@pytest.mark.parametrize('specifier', ['~=1.0', '!=1', '<=1', '>=1', '<1', '>1', '===1'])
def test_specifier_unsupported(specifier: str) -> None:
    with pytest.raises(ValueError, match=r'.*operator not supported:.*'):
        core.version_specifier_condition(Specifier(specifier))



import calendar
import json
import os
from datetime import date, datetime
from typing import Any, Dict, Iterable, List, Optional, Tuple

from google.cloud.bigquery import Client
from google.cloud.bigquery.job import QueryJobConfig
from google.cloud.bigquery.table import RowIterator
from packaging.utils import canonicalize_name
from packaging.requirements import Requirement
from packaging.specifiers import SpecifierSet, Specifier
from packaging.version import Version

from pypinfo.fields import AGGREGATES, Downloads, Field

FROM = 'FROM `bigquery-public-data.pypi.file_downloads`'
DATE_ADD = 'TIMESTAMP_ADD(CURRENT_TIMESTAMP(), INTERVAL {} DAY)'
START_TIMESTAMP = 'TIMESTAMP("{} 00:00:00")'
END_TIMESTAMP = 'TIMESTAMP("{} 23:59:59")'
START_DATE = '-31'
END_DATE = '-1'
DEFAULT_LIMIT = 10

Rows = List[List[str]]


def create_config() -> QueryJobConfig:
    config = QueryJobConfig()
    config.use_legacy_sql = False
    return config


def normalize_dates(start_date: str, end_date: str) -> Tuple[str, str]:
    """If a date is yyyy-mm, normalize as first or last yyyy-mm-dd of the month.
    Otherwise, return unchanged.
    """
    try:
        start_date, _ = month_ends(start_date)  # yyyy-mm
    except (AttributeError, ValueError):
        pass  # -n, yyyy-mm-dd

    try:
        _, end_date = month_ends(end_date)  # yyyy-mm
    except (AttributeError, ValueError):
        pass  # -n, yyyy-mm-dd

    return start_date, end_date


def create_client(creds_file: Optional[str] = None) -> Client:
    creds_file = creds_file or os.environ.get('GOOGLE_APPLICATION_CREDENTIALS')

    if creds_file is None:
        raise SystemError('Credentials could not be found.')

    with open(creds_file, encoding="UTF-8") as file:
        creds = json.load(file)
    project = creds['project_id']
    return Client.from_service_account_json(creds_file, project=project)


def validate_date(date_text: str) -> bool:
    """Return True if valid, raise ValueError if not"""
    try:
        if int(date_text) < 0:
            return True
    except ValueError:
        pass

    try:
        datetime.strptime(date_text, '%Y-%m-%d')
        return True
    except ValueError:
        pass

    raise ValueError('Dates must be negative integers or YYYY-MM[-DD] in the past.')


def format_date(date: str, timestamp_format: str) -> str:
    try:
        date = DATE_ADD.format(int(date))
    except ValueError:
        date = timestamp_format.format(date)
    return date


def month_ends(yyyy_mm: str) -> Tuple[str, str]:
    """Helper to return start_date and end_date of a month as yyyy-mm-dd"""
    year, month = map(int, yyyy_mm.split("-"))
    first = date(year, month, 1)
    number_of_days = calendar.monthrange(year, month)[1]
    last = date(year, month, number_of_days)
    return str(first), str(last)


def strip_trailing_zero(release: Tuple[int, ...]) -> Tuple[int, ...]:
    """Helper to strip trailing 0 in a tuple of integers"""
    new_len = len(release)
    while new_len > 1 and release[new_len - 1] == 0:
        new_len -= 1
    return release[:new_len]


def version_specifier_condition(specifier: Specifier) -> str:
    if specifier.operator == '==':
        if specifier.version.endswith(".*"):
            version_ = Version(specifier.version[:-2])
            prefix_match = True
        else:
            version_ = Version(specifier.version)
            prefix_match = False
        # ignore case
        regex = '(?i)^'
        if version_.epoch == 0:
            regex += '(0+!)?'
        else:
            regex += f'0*{version_.epoch}!'
        exact_release = r'\.'.join(f'0*{i}' for i in strip_trailing_zero(version_.release))
        exact_release += r'(\.0+)*'
        if prefix_match:
            regex += '('
            regex += exact_release
            regex += '|'
            regex += r'\.'.join(f'0*{i}' for i in version_.release)
            regex += r'(\.[0-9]+)*'
            regex += ')'
            # pre
            regex += r'([-_\.]?(a|b|c|rc|alpha|beta|pre|preview)[-_\.]?[0-9]+?)?'
            # post
            regex += r'(-[0-9]+|([-_\.]?(post|rev|r)[-_\.]?[0-9]+?))?'
            # dev
            regex += r'([-_\.]?dev[-_\.]?[0-9]+?)?'
        else:
            regex += exact_release
            if version_.pre is not None:
                regex += r'[-_\.]?'
                regex += {
                    'a': '(a|alpha)',
                    'b': '(b|beta)',
                    'rc': '(c|rc|pre|preview)',
                }[version_.pre[0]]
                if version_.pre[1] == 0:
                    regex += r'([-_\.]?0+)?'
                else:
                    regex += rf'[-_\.]?0*{version_.pre[1]}'
            if version_.post is not None:
                regex += f'(-0*{version_.post}|'
                regex += r'[-_\.]?(post|rev|r)'
                if version_.post == 0:
                    regex += r'([-_\.]?0+)?'
                else:
                    regex += rf'[-_\.]?0*{version_.post}'
                regex += ')'
            if version_.dev is not None:
                regex += r'[-_\.]?dev'
                if version_.dev == 0:
                    regex += r'([-_\.]?0+)?'
                else:
                    regex += rf'[-_\.]?0*{version_.dev}'
        regex += '$'
        return f'REGEXP_CONTAINS(file.version, r"{regex}")\n'

    raise ValueError(f'operator not supported: {specifier}')


def build_query(
    project: str,
    all_fields: Iterable[Field],
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    days: Optional[int] = None,
    limit: Optional[int] = None,
    where: Optional[str] = None,
    order: Optional[str] = None,
    pip: bool = False,
) -> str:
    project_specifiers = SpecifierSet()
    if project:
        req = Requirement(project)
        if req.extras:
            raise ValueError("project: extras are not allowed")
        if req.marker:
            raise ValueError("project: marker is not allowed")
        if req.url:
            raise ValueError("project: url is not allowed")
        project = req.name
        project_specifiers = req.specifier

    project = canonicalize_name(project)

    start_date = start_date or START_DATE
    end_date = end_date or END_DATE
    limit = limit or DEFAULT_LIMIT

    if days:
        start_date = str(int(end_date) - days)

    start_date, end_date = normalize_dates(start_date, end_date)
    validate_date(start_date)
    validate_date(end_date)

    try:
        int(start_date)
        int(end_date)
    except ValueError:
        # Not integers, must be yyyy-mm-dd
        pass
    else:
        if int(start_date) >= int(end_date):
            raise ValueError('End date must be greater than start date.')

    start_date = format_date(start_date, START_TIMESTAMP)
    end_date = format_date(end_date, END_TIMESTAMP)

    # deduplicate fields keeping first occurrence
    # python 3.7+ dict are ordered, keys can be used as an ordered set
    fields = list(dict.fromkeys(all_fields))

    fields.append(Downloads)
    query = 'SELECT\n'

    for field in fields:
        query += f'  {field.data} as {field.name},\n'

    query += f'{FROM}\n'

    conditions = [f'WHERE timestamp BETWEEN {start_date} AND {end_date}\n']
    if project:
        conditions.append(f'file.project = "{project}"\n')
    for specifier in project_specifiers:
        conditions.append(version_specifier_condition(Specifier(str(specifier))))
    if pip:
        conditions.append('details.installer.name = "pip"\n')
    query += '  AND '.join(conditions)
    if where:
        query += f'  AND {where}\n'

    non_aggregate_fields = [field.name for field in fields if field not in AGGREGATES]
    if non_aggregate_fields:
        query += 'GROUP BY\n'
        query += '  '
        query += ', '.join(non_aggregate_fields)
        query += '\n'

    query += f'ORDER BY\n  {order or Downloads.name} DESC\n'
    query += f'LIMIT {limit}'

    return query


def parse_query_result(query_rows: RowIterator) -> Rows:
    rows = [[field.name for field in query_rows.schema]]
    rows.extend([str(item) for item in row] for row in query_rows)
    return rows


def add_percentages(rows: Rows, include_sign: bool = True) -> Rows:

    headers = rows.pop(0)
    index = headers.index('download_count')
    headers.insert(index, 'percent')
    total_downloads = sum(int(row[index]) for row in rows)
    percent_format = '{:.2%}' if include_sign else '{:.2}'

    for row in rows:
        percent = percent_format.format(int(row[index]) / total_downloads)
        row.insert(index, percent)

    rows.insert(0, headers)
    return rows


def get_download_total(rows: Rows) -> Tuple[int, int]:
    """Return the total downloads, and the downloads column"""
    headers = rows.pop(0)
    index = headers.index('download_count')
    total_downloads = sum(int(row[index]) for row in rows)

    rows.insert(0, headers)
    return total_downloads, index


def add_download_total(rows: Rows) -> Rows:
    """Add a final row to rows showing the total downloads"""
    total_row = [""] * len(rows[0])
    total_row[0] = "Total"
    total_downloads, downloads_column = get_download_total(rows)
    total_row[downloads_column] = str(total_downloads)
    rows.append(total_row)

    return rows


def tabulate(rows: Rows, markdown: bool = False) -> str:
    column_widths = [0] * len(rows[0])
    right_align = [[False] * len(rows[0])] * len(rows)

    # Get max width of each column
    for r, row in enumerate(rows):
        for i, item in enumerate(row):
            if item.isdigit():
                # Separate the thousands
                rows[r][i] = f"{int(item):,}"
                right_align[r][i] = True
            elif item.endswith('%'):
                right_align[r][i] = True
            length = len(item)
            if length > column_widths[i]:
                column_widths[i] = length

    tabulated = '| '

    headers = rows.pop(0)
    for i, item in enumerate(headers):
        tabulated += item + ' ' * (column_widths[i] - len(item) + 1) + '| '

    tabulated = tabulated.rstrip()
    tabulated += '\n| '

    for i in range(len(rows[0])):
        tabulated += '-' * (column_widths[i] - 1)
        if right_align[0][i] and markdown:
            tabulated += ': | '
        else:
            tabulated += '- | '

    tabulated = tabulated.rstrip()
    tabulated += '\n'

    for r, row in enumerate(rows):
        for i, item in enumerate(row):
            num_spaces = column_widths[i] - len(item)
            tabulated += '| '
            if right_align[r][i]:
                tabulated += ' ' * num_spaces + item + ' '
            else:
                tabulated += item + ' ' * (num_spaces + 1)
        tabulated += '|\n'

    return tabulated


def format_json(rows: Rows, query_info: Dict[str, Any], indent: Optional[int]) -> str:
    headers, *_data = rows
    data: List[Any] = _data
    items: List[Dict[str, Any]] = []

    for d in data:
        item: Dict[str, Any] = {}
        for i in range(len(headers)):
            if d[i].isdigit():
                d[i] = int(d[i])
            item[headers[i]] = d[i]
        items.append(item)

    now = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')

    j = {'last_update': now, 'rows': items, 'query': query_info}

    separators = (',', ':') if indent is None else None
    return json.dumps(j, indent=indent, separators=separators, sort_keys=True)



import contextlib
import os
from typing import Iterator, Optional

from platformdirs import user_data_dir
from tinydb import TinyDB, Query, where
from tinydb.table import Table
from tinyrecord import transaction

DB_FILE = os.path.join(user_data_dir('pypinfo', appauthor=False), 'db.json')


@contextlib.contextmanager
def get_credentials_table(table: Optional[Table] = None) -> Iterator[Table]:
    if table is None:
        with TinyDB(DB_FILE, create_dirs=True) as db:
            yield db.table('credentials')
    else:
        yield table


def get_credentials(table: Optional[Table] = None) -> Optional[str]:
    with get_credentials_table(table) as table:
        query = table.search(Query().path.exists())
        return query[0]['path'] if query else None


def set_credentials(creds_file: str) -> None:
    with get_credentials_table() as table:
        exists = get_credentials(table)
        with transaction(table) as tr:
            if exists:
                tr.update({'path': creds_file}, where('path').exists())
            else:
                tr.insert({'path': creds_file})



from pathlib import Path

from pypinfo import db

CREDS_FILE = '/path/to/creds_file.json'


def test_get_credentials(tmp_path: Path) -> None:
    # Arrange
    db.DB_FILE = str(tmp_path / 'db.json')  # Mock

    # Assert
    assert db.get_credentials() is None


def test_set_credentials(tmp_path: Path) -> None:
    # Arrange
    db.DB_FILE = str(tmp_path / 'db.json')  # Mock

    # Act
    db.set_credentials(CREDS_FILE)


def test_set_credentials_twice(tmp_path: Path) -> None:
    # Arrange
    db.DB_FILE = str(tmp_path / 'db.json')  # Mock

    # Act
    db.set_credentials(CREDS_FILE)
    db.set_credentials(CREDS_FILE)


def test_round_trip(tmp_path: Path) -> None:
    # Arrange
    db.DB_FILE = str(tmp_path / 'db.json')  # Mock

    # Act
    db.set_credentials(CREDS_FILE)

    # Assert
    assert db.get_credentials() == CREDS_FILE


def test_get_credentials_table(tmp_path: Path) -> None:
    db.DB_FILE = str(tmp_path / 'db.json')
    with db.get_credentials_table() as table:
        assert not table._storage._handle.closed
        with db.get_credentials_table(table) as table2:
            assert table2 is table
        assert not table._storage._handle.closed
    assert table._storage._handle.closed



from decimal import ROUND_UP, Decimal
from typing import List

import click
from binary import TEBIBYTE, convert_units

from pypinfo.core import (
    add_percentages,
    add_download_total,
    build_query,
    create_client,
    create_config,
    format_json,
    month_ends,
    parse_query_result,
    tabulate,
)
from pypinfo.db import get_credentials, set_credentials
from pypinfo.fields import (
    Project,
    Date,
    Month,
    Year,
    Country,
    Version,
    File,
    PythonVersion,
    Percent3,
    Percent2,
    Installer,
    InstallerVersion,
    SetuptoolsVersion,
    System,
    SystemRelease,
    Implementation,
    ImplementationVersion,
    OpenSSLVersion,
    Distro,
    DistroVersion,
    CI,
    CPU,
    Libc,
    LibcVersion,
)

CONTEXT_SETTINGS = {
    'help_option_names': ('-h', '--help'),
    'max_content_width': 300,
}
FIELD_MAP = {
    'project': Project,
    'version': Version,
    'file': File,
    'pyversion': PythonVersion,
    'percent3': Percent3,
    'percent2': Percent2,
    'impl': Implementation,
    'impl-version': ImplementationVersion,
    'openssl': OpenSSLVersion,
    'date': Date,
    'month': Month,
    'year': Year,
    'country': Country,
    'installer': Installer,
    'installer-version': InstallerVersion,
    'setuptools-version': SetuptoolsVersion,
    'system': System,
    'system-release': SystemRelease,
    'distro': Distro,
    'distro-version': DistroVersion,
    'ci': CI,
    'cpu': CPU,
    'libc': Libc,
    'libc-version': LibcVersion,
}
TIER_COST = 5
TB = Decimal(TEBIBYTE)
TO_CENTS = Decimal('0.00')


@click.group(invoke_without_command=True, context_settings=CONTEXT_SETTINGS)
@click.argument('project', required=False)
@click.argument('fields', nargs=-1, required=False)
@click.option('--auth', '-a', help='Path to Google credentials JSON file.')
@click.option('--run/--test', default=True, help='--test simply prints the query.')
@click.option('--json', '-j', is_flag=True, help='Print data as JSON, with keys `rows` and `query`.')
@click.option('--indent', '-i', type=int, help='JSON indentation level.')
@click.option('--timeout', '-t', type=int, default=120000, help='Milliseconds. Default: 120000 (2 minutes)')
@click.option('--limit', '-l', type=int, help='Maximum number of query results. Default: 10')
@click.option('--days', '-d', type=int, help='Number of days in the past to include. Default: 30')
@click.option('--start-date', '-sd', help='Must be negative or YYYY-MM[-DD]. Default: -31')
@click.option('--end-date', '-ed', help='Must be negative or YYYY-MM[-DD]. Default: -1')
@click.option('--month', '-m', help='Shortcut for -sd & -ed for a single YYYY-MM month.')
@click.option('--where', '-w', help='WHERE conditional. Default: file.project = "project"')
@click.option('--order', '-o', help='Field to order by. Default: download_count')
@click.option('--all', 'all_installers', is_flag=True, help='Show downloads by all installers, not only pip.')
@click.option('--percent', '-pc', is_flag=True, help='Print percentages.')
@click.option('--markdown', '-md', is_flag=True, help='Output as Markdown.')
@click.option('--verbose', '-v', is_flag=True, help='Print debug messages to stderr.')
@click.version_option()
@click.pass_context
def pypinfo(
    ctx: click.Context,
    project: str,
    fields: List[str],
    auth: str,
    run: bool,
    json: bool,
    indent: int,
    timeout: int,
    limit: int,
    days: int,
    start_date: str,
    end_date: str,
    month: str,
    where: str,
    order: str,
    all_installers: bool,
    percent: bool,
    markdown: bool,
    verbose: bool,
) -> None:
    """Valid fields are:\n
    project | version | file | pyversion | percent3 | percent2 | impl | impl-version |\n
    openssl | date | month | year | country | installer | installer-version |\n
    setuptools-version | system | system-release | distro | distro-version | ci | cpu |\n
    libc | libc-version
    """
    if auth:
        set_credentials(auth)
        click.echo(f'Credentials location set to "{get_credentials()}".')
        return

    if verbose:
        click.echo(f'Credentials location set to "{get_credentials()}".', err=True)

    if project is None and not fields:
        click.echo(ctx.get_help())
        return

    parsed_fields = []
    for field in fields:
        parsed = FIELD_MAP.get(field)
        if parsed is None:
            raise ValueError(f'"{field}" is an unsupported field.')
        parsed_fields.append(parsed)

    order_name = order
    parsed_order = FIELD_MAP.get(order)
    if parsed_order:
        order_name = parsed_order.name
        parsed_fields.insert(0, parsed_order)

    if month:
        start_date, end_date = month_ends(month)

    built_query = build_query(
        project,
        parsed_fields,
        limit=limit,
        days=days,
        start_date=start_date,
        end_date=end_date,
        where=where,
        order=order_name,
        pip=not all_installers,
    )

    if run:
        with create_client(get_credentials()) as client:
            query_job = client.query(built_query, job_config=create_config())
            query_rows = query_job.result(timeout=timeout // 1000)
            rows = parse_query_result(query_rows)

        # Cached
        from_cache = not not query_job.cache_hit

        # Processed
        bytes_processed = query_job.total_bytes_processed or 0
        processed_amount, processed_unit = convert_units(bytes_processed)

        # Billed
        bytes_billed = query_job.total_bytes_billed or 0
        billed_amount, billed_unit = convert_units(bytes_billed)

        # Cost
        billing_tier = query_job.billing_tier or 1
        estimated_cost = Decimal(TIER_COST * billing_tier) / TB * Decimal(bytes_billed)
        estimated_cost_str = str(estimated_cost.quantize(TO_CENTS, rounding=ROUND_UP))

        if len(rows) == 1 and not json:
            # Only headers returned
            click.echo("No data returned, check project name")
            return

        if percent:
            rows = add_percentages(rows, include_sign=not json)

        # Only for tables, and if more than the header row + a single data row
        if len(rows) > 2 and not json:
            rows = add_download_total(rows)

        if not json:
            click.echo(f'Served from cache: {from_cache}')
            click.echo(f'Data processed: {processed_amount:.2f} {processed_unit}')
            click.echo(f'Data billed: {billed_amount:.2f} {billed_unit}')
            click.echo(f'Estimated cost: ${estimated_cost_str}')

            click.echo()
            click.echo(tabulate(rows, markdown))
        else:
            query_info = {
                'cached': from_cache,
                'bytes_processed': bytes_processed,
                'bytes_billed': bytes_billed,
                'estimated_cost': estimated_cost_str,
            }
            click.echo(format_json(rows, query_info, indent))
    else:
        click.echo(built_query)
